 function _nullishCoalesce(lhs, rhsFn) { if (lhs != null) { return lhs; } else { return rhsFn(); } }// 

import {
  asBoolean,
  asEither,
  asNumber,
  asObject,
  asOptional,
  asString
} from 'cleaners'
import {








  SwapAboveLimitError,
  SwapBelowLimitError,
  SwapCurrencyError,
  SwapPermissionError
} from 'edge-core-js/types'

import {

  checkInvalidCodes,
  ensureInFuture,
  makeSwapPluginQuote,
  safeCurrencyCodes
} from '../swap-helpers.js'

// Invalid currency codes should *not* have transcribed codes
// because currency codes with transcribed versions are NOT invalid
const CURRENCY_CODE_TRANSCRIPTION = {
  // Edge currencyCode: exchangeCurrencyCode
  ETH: {
    USDT: 'usdtErc20'
  },
  ZEC: {
    ZEC: 'zaddr'
  }
}
const INVALID_CURRENCY_CODES = {
  from: {
    ETH: ['FTM', 'MATIC'],
    FTM: 'allCodes',
    MATIC: 'allTokens'
  },
  to: {
    ETH: ['FTM', 'MATIC'],
    FTM: 'allCodes',
    MATIC: 'allTokens'
  }
}
const SIDESHIFT_BASE_URL = 'https://sideshift.ai/api/v1'
const pluginId = 'sideshift'
const swapInfo = {
  pluginId,
  displayName: 'SideShift.ai',
  supportEmail: 'help@sideshift.ai'
}
const ORDER_STATUS_URL = 'https://sideshift.ai/orders/'

async function getAddress(
  wallet,
  currencyCode
) {
  const addressInfo = await wallet.getReceiveAddress({ currencyCode })
  return _nullishCoalesce(addressInfo.segwitAddress, () => ( addressInfo.publicAddress))
}

async function checkQuoteError(
  rate,
  request,
  quoteErrorMessage
) {
  const { fromCurrencyCode, fromWallet, toCurrencyCode } = request

  if (quoteErrorMessage === 'Amount too low') {
    const nativeMin = await fromWallet.denominationToNative(
      rate.min,
      fromCurrencyCode
    )
    throw new SwapBelowLimitError(swapInfo, nativeMin)
  }

  if (quoteErrorMessage === 'Amount too high') {
    const nativeMax = await fromWallet.denominationToNative(
      rate.max,
      fromCurrencyCode
    )
    throw new SwapAboveLimitError(swapInfo, nativeMax)
  }

  if (
    /method/i.test(quoteErrorMessage) &&
    /disabled/i.test(quoteErrorMessage)
  ) {
    throw new SwapCurrencyError(swapInfo, fromCurrencyCode, toCurrencyCode)
  }

  if (/country-blocked/i.test(quoteErrorMessage)) {
    throw new SwapPermissionError(swapInfo, 'geoRestriction')
  }
  throw new Error(`SideShift.ai error ${quoteErrorMessage}`)
}

const createSideshiftApi = (baseUrl, fetch) => {
  async function request(
    method,
    path,
    body
  ) {
    const url = `${baseUrl}${path}`

    const reply = await (method === 'GET'
      ? fetch(url)
      : fetch(url, {
          method,
          headers: {
            'Content-Type': 'application/json'
          },
          body: JSON.stringify(body)
        }))

    try {
      return await reply.json()
    } catch (e) {
      throw new Error(`SideShift.ai returned error code ${reply.status}`)
    }
  }

  return {
    get: (path) => request('GET', path),
    post: (path, body) =>
      request('POST', path, body)
  }
}

const createFetchSwapQuote = (api, affiliateId) =>
  async function fetchSwapQuote(
    request
  ) {
    checkInvalidCodes(INVALID_CURRENCY_CODES, request, swapInfo)

    const [depositAddress, settleAddress] = await Promise.all([
      getAddress(request.fromWallet, request.fromCurrencyCode),
      getAddress(request.toWallet, request.toCurrencyCode)
    ])

    const { safeFromCurrencyCode, safeToCurrencyCode } = safeCurrencyCodes(
      CURRENCY_CODE_TRANSCRIPTION,
      request
    )

    const rate = asRate(
      await api.get(
        `/pairs/${safeFromCurrencyCode}/${safeToCurrencyCode}`
      )
    )

    if (rate.error) {
      throw new SwapCurrencyError(
        swapInfo,
        request.fromCurrencyCode,
        request.toCurrencyCode
      )
    }

    const permissions = asPermissions(await api.get('/permissions'))

    if (!permissions.createOrder || !permissions.createQuote) {
      throw new SwapPermissionError(swapInfo, 'geoRestriction')
    }

    const quoteAmount = await (request.quoteFor === 'from'
      ? request.fromWallet.nativeToDenomination(
          request.nativeAmount,
          request.fromCurrencyCode
        )
      : request.toWallet.nativeToDenomination(
          request.nativeAmount,
          request.toCurrencyCode
        ))

    const depositAmount =
      request.quoteFor === 'from'
        ? quoteAmount
        : (parseFloat(quoteAmount) / parseFloat(rate.rate))
            .toFixed(8)
            .toString()

    const fixedQuoteRequest = asFixedQuoteRequest({
      depositMethod: safeFromCurrencyCode,
      settleMethod: safeToCurrencyCode,
      depositAmount
    })

    const fixedQuote = asFixedQuote(
      await api.post('/quotes', fixedQuoteRequest)
    )

    if (fixedQuote.error) {
      await checkQuoteError(rate, request, fixedQuote.error.message)
      throw new Error(`SideShift.ai error ${fixedQuote.error.message}`)
    }

    const orderRequest = asOrderRequest({
      type: 'fixed',
      quoteId: fixedQuote.id,
      affiliateId,
      settleAddress,
      refundAddress: depositAddress
    })

    const order = asOrder(
      await api.post('/orders', orderRequest)
    )

    if (order.error) {
      await checkQuoteError(rate, request, order.error.message)
      throw new Error(`SideShift.ai error ${order.error.message}`)
    }

    const spendInfoAmount = await request.fromWallet.denominationToNative(
      order.depositAmount,
      request.fromCurrencyCode.toUpperCase()
    )

    const amountExpectedFromNative = await request.fromWallet.denominationToNative(
      order.depositAmount,
      request.fromCurrencyCode
    )

    const amountExpectedToNative = await request.fromWallet.denominationToNative(
      order.settleAmount,
      request.toCurrencyCode
    )

    const isEstimate = false

    const destinationTag =
      order.depositAddress.destinationTag != null
        ? `${order.depositAddress.destinationTag}`
        : undefined

    const spendInfo = {
      currencyCode: request.fromCurrencyCode,
      spendTargets: [
        {
          nativeAmount: spendInfoAmount,
          publicAddress: order.depositAddress.address,
          uniqueIdentifier: _nullishCoalesce(order.depositAddress.memo, () => ( destinationTag))
        }
      ],
      networkFeeOption:
        request.fromCurrencyCode.toUpperCase() === 'BTC' ? 'high' : 'standard',
      swapData: {
        orderId: order.orderId,
        orderUri: ORDER_STATUS_URL + order.orderId,
        isEstimate,
        payoutAddress: settleAddress,
        payoutCurrencyCode: request.toCurrencyCode,
        payoutNativeAmount: amountExpectedToNative,
        payoutWalletId: request.toWallet.id,
        plugin: { ...swapInfo },
        refundAddress: depositAddress
      }
    }

    const tx = await request.fromWallet.makeSpend(spendInfo)

    return makeSwapPluginQuote(
      request,
      amountExpectedFromNative,
      amountExpectedToNative,
      tx,
      settleAddress,
      pluginId,
      isEstimate,
      ensureInFuture(new Date(order.expiresAtISO)),
      order.id
    )
  }

export function makeSideshiftPlugin(
  opts
) {
  const { io, initOptions } = opts

  const api = createSideshiftApi(SIDESHIFT_BASE_URL, io.fetchCors || io.fetch)

  const fetchSwapQuote = createFetchSwapQuote(api, initOptions.affiliateId)

  return {
    swapInfo,
    fetchSwapQuote
  }
}

















const asError = asObject({ error: asObject({ message: asString }) })

const asPermissions = asObject({
  createOrder: asBoolean,
  createQuote: asBoolean
})

const asRate = asEither(
  asObject({
    rate: asString,
    min: asString,
    max: asString
  }),
  asError
)

const asFixedQuoteRequest = asObject({
  depositMethod: asString,
  settleMethod: asString,
  depositAmount: asString
})

const asFixedQuote = asEither(
  asObject({
    id: asString
  }),
  asError
)

const asOrderRequest = asObject({
  type: asString,
  quoteId: asString,
  affiliateId: asString,
  settleAddress: asString,
  refundAddress: asString
})

const asOrder = asEither(
  asObject({
    expiresAtISO: asString,
    depositAddress: asObject({
      address: asString,
      memo: asOptional(asString),
      destinationTag: asOptional(asNumber)
    }),
    id: asString,
    orderId: asString,
    settleAmount: asString,
    depositAmount: asString
  }),
  asError
)
